"""This module defines functions that should be importable for integration tests and unit tests as well.

We use a separate module from conftest.py to circumvent imports from conftest, since this name can be ambiguous
when multiple conftest.py exist (e.g. for unit and integration tests) in one run.
Not importing from conftest is a best practice described in the note here:
https://pytest.org/en/6.2.x/writing_plugins.html#conftest-py-local-per-directory-plugins

This means we only use conftest.py to define fixtures which will automatically be shared with the respective scope.
"""
import os
from collections import namedtuple
from contextlib import AbstractContextManager
from dataclasses import field
from types import TracebackType
from typing import Any, List, Optional, Type

import pynvml
from mockito import when

from scaffold.conf.scaffold.entrypoint import EntrypointConf
from scaffold.entrypoints.entrypoint import Entrypoint
from scaffold.hydra.config_helpers import structured_config

MB = 1024 * 1024


class TmpCwd(object):
    def __init__(self, tmpdir: str) -> None:
        """Jump to a different cwd within the with clause"""
        self.tmpdir = tmpdir

    def __enter__(self) -> "TmpCwd":
        """Saves current cwd and chdir to tmp cwd"""
        self._cwd_stack = os.getcwd()
        os.chdir(self.tmpdir)
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        """Restores stored cwd"""
        os.chdir(self._cwd_stack)


@structured_config(group="my_project/schemas")
class ExampleEntrypointConf(EntrypointConf):
    defaults: List[Any] = field(
        default_factory=lambda: [
            "/scaffold/entrypoint/hydra_defaults@_global_",
            {"/scaffold/entrypoint/logging@logging": "default"},
        ]
    )
    greeting: str = "Hey"


class MyContext(AbstractContextManager):
    """Helper context manager for testing. It sets a flag True when active and False when it exits."""

    def __init__(self) -> None:
        """Context init"""
        self.ctx_active = None

    def __enter__(self) -> "MyContext":
        """Context enter"""
        self.ctx_active = True
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        """Context exit"""
        self.ctx_active = False


class MinimalExampleEntrypoint(Entrypoint[ExampleEntrypointConf]):
    def run(self, name: str) -> str:
        """Compile and return greeting."""
        return self.config.greeting + " " + name


def configure_mock(N: object = pynvml, scenario_nonexistent_pid: bool = False) -> None:
    """Define mock behaviour for pynvml and psutil.{Process,virtual_memory}.

    Functionality adopted from:
    https://github.com/wookayin/gpustat/blob/cab5f2465fe718619ba03bf59f5208909581ea60/gpustat/test_gpustat.py#L202
    """

    # without following patch, unhashable NVMLError makes unit test crash
    N.NVMLError.__hash__ = lambda _: 0
    assert issubclass(N.NVMLError, BaseException)

    when(N).nvmlInit().thenReturn()
    when(N).nvmlShutdown().thenReturn()
    when(N).nvmlSystemGetDriverVersion().thenReturn("415.27.mock")

    NUM_GPUS = 3
    mock_handles = ["mock-handle-%d" % i for i in range(3)]
    when(N).nvmlDeviceGetCount().thenReturn(NUM_GPUS)

    def _return_or_raise(v: Any) -> Any:
        """Return a callable for thenAnswer() to let exceptions re-raised."""

        def _callable(*args, **kwargs) -> Any:
            if isinstance(v, Exception):
                raise v
            return v

        return _callable

    for i in range(NUM_GPUS):
        handle = mock_handles[i]
        when(N).nvmlDeviceGetHandleByIndex(i).thenReturn(handle)
        when(N).nvmlDeviceGetIndex(handle).thenReturn(i)
        when(N).nvmlDeviceGetName(handle).thenReturn(("GeForce GTX TITAN %d" % i).encode())
        when(N).nvmlDeviceGetUUID(handle).thenReturn(
            {
                0: b"GPU-10fb0fbd-2696-43f3-467f-d280d906a107",
                1: b"GPU-d1df4664-bb44-189c-7ad0-ab86c8cb30e2",
                2: b"GPU-50205d95-57b6-f541-2bcb-86c09afed564",
            }[i]
        )

        when(N).nvmlDeviceGetTemperature(handle, N.NVML_TEMPERATURE_GPU).thenReturn([80, 36, 71][i])
        when(N).nvmlDeviceGetFanSpeed(handle).thenReturn([16, 53, 100][i])
        when(N).nvmlDeviceGetPowerUsage(handle).thenAnswer(
            _return_or_raise({0: 125000, 1: N.NVMLError_NotSupported(), 2: 250000}[i])
        )
        when(N).nvmlDeviceGetEnforcedPowerLimit(handle).thenAnswer(
            _return_or_raise({0: 250000, 1: 250000, 2: N.NVMLError_NotSupported()}[i])
        )

        mock_memory_t = namedtuple("Memory_t", ["total", "used"])
        when(N).nvmlDeviceGetMemoryInfo(handle).thenAnswer(
            _return_or_raise(
                {
                    0: mock_memory_t(total=12883853312, used=8000 * MB),
                    1: mock_memory_t(total=12781551616, used=9000 * MB),
                    2: mock_memory_t(total=12781551616, used=0),
                }[i]
            )
        )

        mock_utilization_t = namedtuple("Utilization_t", ["gpu", "memory"])
        when(N).nvmlDeviceGetUtilizationRates(handle).thenAnswer(
            _return_or_raise(
                {
                    0: mock_utilization_t(gpu=76, memory=0),
                    1: mock_utilization_t(gpu=0, memory=0),
                    2: N.NVMLError_NotSupported(),  # Not Supported
                }[i]
            )
        )

        when(N).nvmlDeviceGetEncoderUtilization(handle).thenAnswer(
            _return_or_raise(
                {
                    0: [88, 167000],  # [value, sample_rate]
                    1: [0, 167000],  # [value, sample_rate]
                    2: N.NVMLError_NotSupported(),  # Not Supported
                }[i]
            )
        )
        when(N).nvmlDeviceGetDecoderUtilization(handle).thenAnswer(
            _return_or_raise(
                {
                    0: [67, 167000],  # [value, sample_rate]
                    1: [0, 167000],  # [value, sample_rate]
                    2: N.NVMLError_NotSupported(),  # Not Supported
                }[i]
            )
        )

        # running process information: a bit annoying...
        mock_process_t = namedtuple("Process_t", ["pid", "usedGpuMemory"])

        if scenario_nonexistent_pid:
            mock_processes_gpu2_erratic = [
                mock_process_t(99999, 9999 * MB),
                mock_process_t(99995, 9995 * MB),
            ]
        else:
            mock_processes_gpu2_erratic = N.NVMLError_NotSupported()
        when(N).nvmlDeviceGetComputeRunningProcesses(handle).thenAnswer(
            _return_or_raise(
                {
                    0: [mock_process_t(48448, 4000 * MB), mock_process_t(153223, 4000 * MB)],
                    1: [mock_process_t(192453, 3000 * MB), mock_process_t(194826, 6000 * MB)],
                    2: mock_processes_gpu2_erratic,  # Not Supported or non-existent
                }[i]
            )
        )

        when(N).nvmlDeviceGetGraphicsRunningProcesses(handle).thenAnswer(
            _return_or_raise(
                {
                    0: [mock_process_t(48448, 4000 * MB)],
                    1: [],
                    2: N.NVMLError_NotSupported(),
                }[i]
            )
        )



"""Taken from https://github.com/facebookresearch/hydra/blob/main/examples/plugins/example_searchpath_plugin/hydra_plugins/example_searchpath_plugin/example_searchpath_plugin.py"""  # noqa: E501
from hydra.core.config_search_path import ConfigSearchPath
from hydra.plugins.search_path_plugin import SearchPathPlugin


class ScaffoldSearchPathPlugin(SearchPathPlugin):
    def manipulate_search_path(self, search_path: ConfigSearchPath) -> None:
        """Appends the search path for this plugin to the end of the hydra search path.
        This makes all config files inside the conf folder available by their group/name.
        """
        from scaffold.conf import register_all

        register_all()

        search_path.append(provider="scaffold", path="pkg://scaffold/conf")



# Enables postponed evaluation of type hints. See https://www.python.org/dev/peps/pep-0563/
# This is especially useful when using type hints for modules that we want to import lazily.
from __future__ import annotations

import logging
from copy import deepcopy
from pathlib import Path
from typing import Dict, List, Optional, TYPE_CHECKING

from scaffold.data.artifact_manager.base import ArtifactManager, DirectoryLogger

# Enables the usage for type hints. Will not be executed during runtime, because we want to use lazy imports.
if TYPE_CHECKING:
    import torch

logger = logging.getLogger(__name__)

STATE_FILENAME = "state.pt"


class ModelLogger:
    """Class to be used for logging a model or state dicts as an artifact.
    Will always use a temporary dir locally to save artifacts before logging that dir.
    """

    def __init__(self, artifact_manager: ArtifactManager) -> None:
        """
        Initializes the ModelLogger.

        Args:
            artifact_manager (ArtifactManager): ArtifactManager to use for logging artifacts.
        """
        self.artifact_manager = artifact_manager

    @staticmethod
    def save_state(
        dict_dp: str,
        model: torch.nn.Module,
        optimizers: List[torch.optim.Optimizer] = None,
        schedulers: List[torch.optim.lr_scheduler._LRScheduler] = None,
        **kwargs,
    ) -> None:
        """Saves the model and optimizer state dicts, and optionally some keyword arguments, into one dictionary
        that can later be loaded to retrieve the exact state of the model training.

        Notes:
            We intentionally do not save the model.state_dict() here since we need to support easy
            model loading from a single afid in the evaluator without initializing the model class first.

        Args:
            dict_dp (str): Directory where the state dict pickle file should be saved.
            model (torch.nn.Module): Model to save the state dict off.
            optimizers (List[torch.optim.Optimizer]): All optimizers to save the state dicts off.
            kwargs: Additional key value pairs to save to the state dict
        """
        import torch

        # If the model is distributed, we want to unwrap it first, so that the checkpoint does not depend on
        # being loaded as a distributed model. When later loading the state, we use the state dict to load
        # the weights, which makes it agnostic towards the exact type of the model.
        state = {
            "model": model.module if isinstance(model, torch.nn.parallel.DistributedDataParallel) else model,
            **kwargs,
        }
        if optimizers is not None:
            state.update({"optimizer_state_dicts": [o.state_dict() for o in optimizers]})
        if schedulers is not None:
            scheduler_state_dicts = []
            for s in schedulers:
                # Lightning wraps the schedulers with extra parameters we would also like to save (e.g. "interval").
                # We overwrite the key "scheduler" where the actual class lives, with its state dict, and keep the rest.
                s = deepcopy(s)
                s["scheduler"] = s["scheduler"].state_dict()
                scheduler_state_dicts.append(s)
            state.update({"scheduler_state_dicts": scheduler_state_dicts})

        dict_dp = Path(dict_dp)
        dict_dp.mkdir(exist_ok=True)
        torch.save(state, dict_dp / STATE_FILENAME)

    def log_state_to_artifact(
        self,
        afid: str,
        model: torch.nn.Module,
        optimizers: List[torch.optim.Optimizer] = None,
        collection: Optional[str] = None,
        **kwargs,
    ) -> str:
        """Logs the model and optimizer state dicts under the specified artifact id.

        Args:
            afid (str): Artifact id to log under on the given artifact store or None to generate a new one.
            model (torch.nn.Module): Model to save the state dict off.
            optimizers (List[torch.optim.Optimizer]): All optimizers to save the state dicts off.
            collection (Optional[str]): Collection to log the artifact to.
            kwargs: Additional key value pairs to save to the state dict
        """
        with DirectoryLogger(self.artifact_manager, afid, collection=collection) as dp:
            self.save_state(Path(dp), model, optimizers, **kwargs)

        return afid

    def retrieve_state_from_artifact(
        self, afid: str, collection: Optional[str] = None, version: int = None, device: Optional[str] = None
    ) -> Dict:
        """
        Returns a dictionary containing the model and optimizer state from the artifact store at the artifact_path.

        Args:
            afid: Artifact id of the model. Must have been logged via :py:meth:`log_state_to_artifact`
            device: Device to load the model to. "cpu", "cuda" or None, which is the default that
                will use the gpu if available

        Returns:
            State dictionary containing the model and optimizer state. See :py:meth:`log_state_to_artifact` for the
            key-value pairs.
        """
        import torch

        if device is None:
            device = "cuda" if torch.cuda.is_available() else "cpu"

        with self.artifact_manager.download_artifact(artifact=afid, collection=collection, version=version) as artdir:
            model = torch.load(Path(artdir) / STATE_FILENAME, map_location=torch.device(device), weights_only=False)
        return model



import importlib
import logging

if importlib.util.find_spec("flytekit") is None:
    # if we cannot import flyte, importing flyte_launcher throws and error and we assume that
    # in that case we do not want to have the plugin available at all.
    # this ensure users who did not install flyte extra and dont need this plugin do not get
    # a confusing error message (which could be ignored anyway)
    logging.getLogger(__name__).warning("Could not import flytekit. Hydra flyte launcher plugin disabled")
    __all__ = []
else:
    # from . import  as flyte_launcher
    from . import _flyte_launcher as flyte_launcher
    from ._flyte_launcher import FlyteLauncher

    __all__ = [flyte_launcher, FlyteLauncher]



import os
import pickle
import random
import time
from itertools import islice
from typing import Any, Callable, Iterable, Iterator, List, Optional

import numpy as np
from tqdm import tqdm

from scaffold.data.constants import SeedType


def _pick(buf: List, rng: random.Random) -> Any:
    """Pick a random element from a list.

    Args:
        buf (List): List of items.
        rng (random.Random): Random number generator to use.

    Returns:
        Any: One random item from `buf`.
    """
    k = rng.randint(0, len(buf) - 1)
    sample = buf[k]
    buf[k] = buf[-1]
    buf.pop()
    return sample


def shuffle_(
    iterable: Iterable,
    bufsize: int = 1000,
    initial: int = 100,
    rng: Optional[random.Random] = None,
    seed: SeedType = None,
) -> Iterator:
    """Shuffle the data in the stream.

    Uses a buffer of size `bufsize`. Shuffling at startup is less random; this is traded off against yielding samples
    quickly.

    Args:
        iterable (Iterable): Iterable to shuffle.
        bufsize (int, optional): Buffer size for shuffling. Defaults to 1000.
        initial (int, optional): Minimum number of elements in the buffer before yielding the first element. Must be
            less than or equal to `bufsize`, otherwise will be set to `bufsize`. Defaults to 100.
        rng (random.Random, optional): Either `random` module or a :py:class:`random.Random` instance. If None,
            a `random.Random()` is used.
        seed (Union[int, float, str, bytes, bytearray, None]): A data input that can be used for `random.seed()`.

    Yields:
        Any: Shuffled items of `iterable`.
    """
    rng = get_random_range(rng, seed)

    iterator = iter(iterable)
    initial = min(initial, bufsize)
    buf = []
    for sample in iterator:
        buf.append(sample)
        if len(buf) < bufsize:
            try:
                buf.append(next(iterator))
            except StopIteration:
                pass
        if len(buf) >= initial:
            yield _pick(buf, rng)
    while len(buf) > 0:
        yield _pick(buf, rng)


def take_(iterable: Iterable, n: int) -> Iterator:
    """Yield the first n elements from the iterable.

    Args:
        iterable (Iterable): Iterable to take from.
        n (int): Number of samples to take.

    Yields:
        Any: First `n` elements of `iterable`. Less elements can be yielded if the iterable does not have enough
            elements.
    """
    yield from islice(iterable, 0, n, 1)


def batched_(
    iterable: Iterable,
    batchsize: int = 20,
    collation_fn: Optional[Callable] = None,
    drop_last_if_not_full: bool = True,
) -> Iterator[List]:
    """Yield batches of the given size.

    Args:
        iterable (Iterable): Iterable to be batched.
        batchsize (int, optional): Target batch size. Defaults to 20.
        collation_fn (Callable, optional): Collation function. Defaults to None.
        drop_last_if_not_full (bool, optional): If the length of the last batch is less than `batchsize`, drop it.
            Defaults to True.

    Yields:
        Batches (i.e. lists) of samples.
    """
    it = iter(iterable)
    while True:
        batch = list(islice(it, batchsize))
        if drop_last_if_not_full and len(batch) < batchsize or len(batch) == 0:
            return
        if collation_fn is not None:
            batch = collation_fn(batch)
        yield batch


def map_(iterable: Iterable, callback: Callable) -> Iterator:
    """Apply the `callback` to each item in the `iterable` and yield the item."""
    for sample in iterable:
        yield callback(sample)


def filter_(iterable: Iterable, predicate: Callable) -> Iterator:
    """Filter items in the `iterable` by the `predicate` callable."""
    for sample in iterable:
        if predicate(sample):
            yield sample


def flatten_(iterables: Iterable[Iterable]) -> Iterator:
    """Iterate over iterables in the stream and yield their items."""
    for item in iterables:
        yield from item


def tqdm_(iterable: Iterable, **kwargs) -> Iterator:
    """Iterate while using tqdm."""
    yield from tqdm(iterable, **kwargs)


def get_random_range(rng: Optional[random.Random] = None, seed: SeedType = None) -> random.Random:
    """
    Returns a random number as range, calculated based on the input `rng` and `seed`.

    Args:
        rng (random.Random, optional): Either `random` module or a :py:class:`random.Random` instance. If None,
            a `random.Random()` is used.
        seed (Union[int, float, str, bytes, bytearray, None]): seed (Optional[int]): An int or other acceptable types
            that works for random.seed(). Will be used to seed `rng`. If None, a unique identifier will be used to seed.
    """
    if rng is None:
        rng = random.Random()
    if seed is None:
        seed = f"{os.getpid()}{time.time()}"
    rng.seed(seed)
    return rng


def getsize(item: Any) -> int:
    """Return estimated size (in terms of bytes) of a python object. Currently use numpy method to calculate size.
    Otherwise, the size is estimated through its pickled size. This is considered a better performant option than e.g.
    `zarr.storage.getsize()`.
    """
    if isinstance(item, np.ndarray):
        return item.nbytes
    return len(pickle.dumps(item))



# Enables postponed evaluation of type hints. See https://www.python.org/dev/peps/pep-0563/
# This is especially useful when using type hints for modules that we want to import lazily.
from __future__ import annotations

import logging
from argparse import Namespace
from typing import Any, Dict, Union

from pytorch_lightning.loggers.logger import Logger
from pytorch_lightning.utilities import rank_zero_only

logger = logging.getLogger(__name__)


class HyperoptLightningLogger(Logger):
    def __init__(self):
        """Lightning wrapper around hyperopt logging."""
        super().__init__()
        from scaffold.flyte.hyperopt.loggers import get_hyperopt_logger

        self.logger = get_hyperopt_logger()

    @property
    def name(self) -> str:
        """The experiment name."""
        return f"Logging with {self.logger.__class__}"

    @property
    def version(self) -> str:
        """The experiment version."""
        return self.logger.get_description()

    @rank_zero_only
    def log_hyperparams(self, params: Union[Dict[str, Any], Namespace]) -> None:
        """
        Logging of hyperparameters for experiment.

        Args:
            params (argparse.Namespace): The parameters to be logged.
        """
        pass

    @rank_zero_only
    def log_metrics(self, metrics: Dict[str, Any], step: int) -> None:
        """
        Logging of metrics.

        Args:
            metrics (Dict): Dictionary of metric names and values to be logged
            step (int): The step counter of optimisation iterations
        """
        for metric, value in metrics.items():
            if metric == self.logger.metric:
                self.logger.log_metric(metric, step, value)
                return
        self.logger.logger.warning(
            f"Could not find relevant hyperopt metric {self.logger.metric} in submitted values:\n {metrics}"
        )

    @rank_zero_only
    def finalize(self, status: str) -> None:
        """
        Finalisation of experiment run.

        Args:
            status (str): The experiment outcome (success, failed, etc.).
        """
        pass



"""This module defines shared fixtures that can be used inside a conftest.py like this

from scaffold.integration_test.shared_fixtures import *  # noqa: F401, F403

These fixtures will then be shared with the tests under the scope of the conftest.py.

This is useful for splitting unit and integration tests and defining separate conftest files.
Specific fixtures for integration or unit tests can then be defined in the respective conftest.
"""



"""This module defines specific fixtures for unit tests. Shared fixtures are defined in shared_fixtures.py.

####################################
Please do not import from this file.
####################################

Not importing from conftest is a best practice described in the note here:
https://pytest.org/en/6.2.x/writing_plugins.html#conftest-py-local-per-directory-plugins
"""

from scaffold.integration_test.shared_fixtures import *  # noqa: F401, F403



from __future__ import annotations

import abc
import logging
from abc import ABC
from contextlib import AbstractContextManager
from copy import deepcopy
from typing import Any, Generic, List, TYPE_CHECKING, TypeVar, Union

import hydra
import omegaconf
from omegaconf import OmegaConf

import scaffold
from scaffold.conf.scaffold.entrypoint import EntrypointConf
from scaffold.ctx_manager import combined_context, LoggingContext, TimerContext

if TYPE_CHECKING:
    from scaffold.hydra.config_helpers import StructuredConfig

logger = logging.getLogger(__name__)

ENTRYPOINT_CONFIG_TYPE = TypeVar("ENTRYPOINT_CONFIG_TYPE")


class Entrypoint(ABC, Generic[ENTRYPOINT_CONFIG_TYPE]):
    def __init__(
        self,
        config: ENTRYPOINT_CONFIG_TYPE,
        contexts: List[AbstractContextManager] = None,
    ) -> None:
        """
        Initialize entrypoint with a configuration of the specified type.
        Given context managers will be appended after the ones are found in the config.

        Args:
            config (EntrypointConf): The entrypoint configuration object of the specified type.
            contexts (List[AbstractContextManager]): Context managers that should be added next to the ones
                specified in the config.
        """
        # We check for correct type explicitly, because
        # specifying bounds for the generic type is not supported by vscode yet.
        assert (
            issubclass(OmegaConf.get_type(config), EntrypointConf)
            or OmegaConf.get_type(config) == EntrypointConf
            # if the config is a DictConfig, we can't check the type
            # this happens when composing the config through scaffold.hydra.compose()
            or OmegaConf.get_type(config) == dict
        ), "Config passed to the Entrypoint must be of class EntrypointConf."

        self.config: ENTRYPOINT_CONFIG_TYPE = deepcopy(config)
        self._initialize_contexts(contexts)
        logger.debug("Config:\n" + OmegaConf.to_yaml(self.config))

    def _initialize_contexts(self, contexts: List[AbstractContextManager] = None) -> None:
        """Initializes contexts from the config and adds them to the list of contexts."""
        cls_name = self.__class__.__name__
        logging_ctx = self._get_logging_context(cls_name=cls_name)
        timer_ctx = TimerContext(module_name=cls_name)

        self.contexts = [logging_ctx, timer_ctx]
        if "contexts" in self.config:
            initialized_contexts = [hydra.utils.instantiate(c) for c in self.config.contexts.values()]
            self.contexts.extend(initialized_contexts)
            logger.debug(f"Added contexts from config: {initialized_contexts}")
        if contexts is not None:
            self.contexts.extend(contexts)

    @classmethod
    def from_config_name_or_class(
        cls,
        config_name_or_class: Union[str, StructuredConfig],
        config_dir: str = None,
        overrides: List[str] = None,
        contexts: List[AbstractContextManager] = None,
    ) -> Entrypoint:
        """Temporarily initialized hydra, composes the config and initialized the entrypoint.

        Args:
            config_name_or_class (str): Name of the config file or EntrypointConf class.
            config_dir (str): Absolute path to the directory that should be added to the searchpath.
            overrides (List[str]): Hydra overrides that get applied during config composition.
            contexts (List[AbstractContextManager]): Contexts that should be added to the ones defined in the config.

        Return:
            Initialized Entrypoint instance.
        """
        _config = scaffold.hydra.compose(config_name_or_class, config_dir=config_dir, overrides=overrides)
        return cls(config=_config, contexts=contexts)

    def _get_logging_context(self, cls_name: str) -> LoggingContext:
        """Creates a logging context with the entrypoint class name as filename."""
        try:
            if self.config.logging.handlers.file.filename is None:
                self.config.logging.handlers.file.filename = f"{cls_name}.log"
        except omegaconf.errors.ConfigAttributeError:
            logger.debug(
                "Tried to overwrite the log filename, but couldn't find 'logging.handlers.file.filename' in the config!"
                "\nThis could be because the logging config is not provided."
                " It's recommended to either use '/scaffold/entrypoint/EntrypointConf@_here_' to apply the default"
                " schema and default values, or manually use '/scaffold/entrypoint/logging@logging: default' "
                "in your defaults list."
            )
        return LoggingContext(self.config.logging, verbose=self.config.verbose)

    @abc.abstractmethod
    def run(self, *args, **kwargs) -> Any:
        """Abstract function that has to be overwritten."""
        pass

    def __call__(self, *args, **kwargs) -> Any:
        """Wraps self.run() within initialized contexts. Arguments and keyword arguments will be passed to run()."""
        with combined_context(*self.contexts) as _:  # TODO extend when need access to context module
            output = self.run(*args, **kwargs)
        return output
